/* client.c -- 
 *  Copyright (c) Distributed Multimedia Computing Laboratory,
 *                The University of Texas at Austin. 1996
 *                All Rights Reserved..
 * Author          : Prashant J. Shenoy
 * Created On      : Mon May  6 19:28:09 1996
 * Last Modified By: Prashant J. Shenoy
 * Last Modified On: Thu Jun 26 16:48:30 1997
 * Update Count    : 121
 * Status          : Unknown, Use with caution!
 */


#include "defns.h"
#include "disk.h"
#include "client.h"
#include "proto.h"

Client *clientTable;
EVENT startSimulation;

/* top level simulation process : starts all other processes */

void
simulate(struct arguments &args, DiskArray &diskArray, Statistics &stats)
{
  extern EVENT startSimulation;
  int i;

  create("Simulate");

  /* _________ initilalize __________ */
  reset_prob(args.seed);
  max_events(diskArray.numDisks*9+100);
  max_histograms(20);
  if((clientTable = (Client *) calloc(args.numClients, sizeof(Client))) == NULL)
    error("Memory allocation failed for clientTable\n", E_MEM);
  initStats(args, diskArray,stats); /* init tables/histograms */

  /* __________ start up csim processes _____________ */
  startSimulation = event("start simulation");
  StatisticsProcess(args,diskArray,stats); /* statistics collection process */

 for(i=0;i < diskArray.numDisks; i++){
   initDisk(diskArray.disk[i]);
   DiskProcess(i,stats);
 }
 /* ________ start the failure process to simulate disk failures */
 FailureProcess(args,diskArray);

 /* ________ start the round timeer  process */
 RoundTimerProcess(); 

  if(!args.emptySystem){ /* all clients arrive at t=0 */
   for(i=0;i< args.numClients; i++){
     clientTable[i].id=i;
     if(i<args.numCMClients)
       initClient(args,clientTable[i],VIDEO);
     else
       initClient(args,clientTable[i],TEXT);
     ClientProcess(i);
   }
   set(startSimulation); /* start simulation */
   hold(args.simulationLength);
  }
  else{
     /* clients arrive and depart at random instants */
     error("Not implemented yet",E_ARGS);
   }
  
}

void
initClient(struct arguments &args, Client &client, FileType type)
{
  client.type = type;
  client.holdTime = args.round;
  client.dataInBuffer = client.framesInBuffer = 0;
  client.mode = args.serverPush ? PUSH : PULL;

  switch(type){
  case TEXT:
    client.file = args.numCMFiles + random_int(0,args.numTextFiles-1);
    client.block = random_int(0, fileTable[ client.file].numBlocks-1);
    client.requestSize = args.requestSize;
    client.holdTime = args.holdTime;
    break;
  case VIDEO:
    client.file = random_int(0,args.numCMFiles-1);
    client.block =  random_int(0, fileTable[ client.file].numBlocks-1);
    client.block =  nextMetaBlock(client.block,client.file);
    client.playfp = fileTable[client.file].inode[client.block].firstFrame;
    client.holdTime = args.round;
    client.framesPerRound = args.dataRate * args.round /1000;
    client.requestSize = !args.serverPush ? args.pullRequestSize : 0;
    break;
  }
}

/* seeks to the next meta block */
int
nextMetaBlock(int block, int file)
{
  int total = fileTable[file].numBlocks;
  while(fileTable[file].inode[(total+block-1) %total ].disk == 
	fileTable[file].inode[block].disk)
    block = (block+1)%total;
  return block;
}

void
ClientProcess(int client)
{
  extern EVENT startSimulation;
  RequestList requestList;

  create("client");

  /* wait for the startSimulation Event here */
  /* wait(startSimulation);*/

  while(1){
    generateRequest(requestList,clientTable[client]);
    #ifdef DEBUG
    printf("reading %d blocks for client %d\n",requestList.numRequests,client);
    #endif
    issueRequest(requestList); /* queue up blocks in the appropriate disk queues */
    hold(getHoldTime(clientTable[client]));
  }
}

double
getHoldTime(Client &client)
{
  double holdTime;
    switch(client.type){
    case VIDEO:
     if(client.mode == PUSH)
       holdTime = client.holdTime;
     else /* pull */
       holdTime = client.framesInBuffer * client.holdTime / client.framesPerRound;
      break;
    case TEXT:
      holdTime = expntl(client.holdTime);
      break;
    }
  return holdTime;
}

void
generateRequest(RequestList &requestList, Client &client)
{
 switch(client.type){
 case VIDEO:
   if(client.mode == PUSH)
     generateCMPushRequest(requestList,client);
   else
     generateCMPullRequest(requestList,client);
   break;
 case TEXT:
   generateTextRequest(requestList,client);
   break;
 }
}

void
generateTextRequest(RequestList &requestList, Client &client)
{
  
  int requestSize,bytes,numBlocks,i;


  requestSize = (int) (expntl(client.requestSize)*1024 + 0.99); /* ceil() */
  bytes = requestSize - client.dataInBuffer;
  if(bytes>0){
    numBlocks = bytes/(1024* fileTable[ client.file ].blockSize);
    if(bytes%(1024* fileTable[ client.file ].blockSize))
      numBlocks++;
    client.dataInBuffer = numBlocks*1024* fileTable[ client.file ].blockSize - bytes;
#ifdef DEBUG_REQGEN
    printf("text request for %d blocks\n",numBlocks);
#endif
    for(i=0;i<numBlocks;i++){
      requestList.request[i].type =  TEXT;
      requestList.request[i].client = client.id;
      requestList.request[i].arrivalTime=clock;
      requestList.request[i].disk = fileTable[ client.file ].inode[ client.block ].disk;
      requestList.request[i].blockNo = fileTable[ client.file ].inode[ client.block ].block;
      requestList.request[i].blockSize = fileTable[ client.file ].inode[ client.block ].size;
      client.block = (client.block+1) % fileTable[ client.file ].numBlocks;
      
    }
    requestList.numRequests = numBlocks;
  }
 else{
   client.dataInBuffer -= requestSize;
   requestList.numRequests = 0;
 }
}


void
generateCMPullRequest(RequestList &requestList, Client &client)
{ 
  int framesRetrieved,numRequests,numBlocks,first,last;

  numRequests = framesRetrieved = 0 ;
  numBlocks =  client.requestSize/ diskArray.blockSize; /* num base blocks to be retrieved*/
  if(client.requestSize%diskArray.blockSize)
    numBlocks++; 

  while(numBlocks>0){
      requestList.request[ numRequests ].type = VIDEO;
      requestList.request[ numRequests ].client = client.id;
      requestList.request[ numRequests ].mode = client.mode;
      requestList.request[ numRequests ].arrivalTime = clock;
      requestList.request[ numRequests ].deadline = clock+diskArray.scheduleParam.round;
      requestList.request[ numRequests ].disk = fileTable[ client.file ].inode[ client.block ].disk;
      requestList.request[ numRequests ].blockNo = fileTable[ client.file ].inode[ client.block ].block;
      requestList.request[ numRequests ].blockSize = fileTable[ client.file ].inode[ client.block ].size;
      
      /* _____ update number of frames retrieved so far ______ */
      first = fileTable[client.file].inode[client.block].firstFrame;
      last  = fileTable[client.file].inode[client.block].lastFrame;
      if(last >= first)
	framesRetrieved += (last - first );
      else /* ____ wrap around in frameTable */
	framesRetrieved += (frameTableSize - (first - last));
      if(fileTable[client.file].inode[client.block].residue > 0)
	framesRetrieved++; 
      client.block = (client.block+1)% fileTable[client.file].numBlocks ;
      if(fileTable[client.file].inode[client.block].residue == 0)
	framesRetrieved++;
#ifdef DEBUG_REQGEN
  printf("request f = %3d d = %3d bl = %4d frames [%4d %4d] framesRet = %3d\n", client.file,requestList.request[numRequests].disk, requestList.request[numRequests].blockNo,first,last,framesRetrieved);
#endif
      numRequests++;
      numBlocks--;
    }
    client.framesInBuffer = framesRetrieved; /* store # frames retrieved in this req */
    requestList.numRequests = numRequests;
}

void
generateCMPushRequest(RequestList &requestList, Client &client)
{
  int framesToBeRetrieved, bytes, i, numRequests, first,last;
  framesToBeRetrieved = client.framesPerRound - client.framesInBuffer;

  numRequests =  0;  
  if(framesToBeRetrieved > 0){
    while(framesToBeRetrieved > 0){
      requestList.request[ numRequests ].type = VIDEO;
      requestList.request[ numRequests ].client = client.id;
      requestList.request[ numRequests ].mode = client.mode;
      requestList.request[ numRequests ].arrivalTime=clock;
      requestList.request[ numRequests ].deadline=clock+timeLeftInRound();
      requestList.request[ numRequests ].disk = fileTable[ client.file ].inode[ client.block ].disk;
      requestList.request[ numRequests ].blockNo = fileTable[ client.file ].inode[ client.block ].block;
      requestList.request[ numRequests ].blockSize = fileTable[ client.file ].inode[ client.block ].size;
      
      /* ____ update framesToBeRetrieved ___ */
      first = fileTable[client.file].inode[client.block].firstFrame;
      last  = fileTable[client.file].inode[client.block].lastFrame;
      if(last >= first)
	framesToBeRetrieved -= (last - first );
      else /* ____ wrap around in frameTable */
	framesToBeRetrieved -= (frameTableSize - (first - last));
      if(fileTable[client.file].inode[client.block].residue > 0)
	framesToBeRetrieved--; 
      client.block = (client.block+1)% fileTable[client.file].numBlocks ;
      if(fileTable[client.file].inode[client.block].residue == 0)
	framesToBeRetrieved--;
#ifdef DEBUG_REQGEN
  printf("request f = %3d d = %3d bl = %4d frames [%4d %4d] framesLeft = %3d\n", client.file,requestList.request[numRequests].disk, requestList.request[numRequests].blockNo,first,last,framesToBeRetrieved);
#endif
      numRequests++;
    }
    client.framesInBuffer = -framesToBeRetrieved; /* store # frames from next round retrieved in this rnd*/
    requestList.numRequests = numRequests;
  }
  else{
    client.framesInBuffer -= client.framesPerRound;
    requestList.numRequests = numRequests;
  }
}



     
void
issueRequest(RequestList &requestList)
{
  int i,disk;

  /* request coalescing : coalesce successive requested blocks as a single req */
  /* restrict to RAID0 since it can interfere with recovery in RAID levels 3,4,5*/
  if(diskArray.RAIDLevel == RAID0) 
    coalesceRequests(requestList);
  registerRequest(requestList); /* register the request with the stats routine */
  for(i=0;i<requestList.numRequests;i++){
    disk = requestList.request[i].disk;
    switch(requestList.request[i].type){
    case VIDEO:
      if(diskArray.disk[disk].status == UP){
	if(requestList.request[i].mode == PUSH)
	  insertIntoQueue(diskArray.disk[disk].periodicQueue,requestList.request[i]);
                  else /* PULL */
                   insertIntoQueue(diskArray.disk[disk].aperiodicQueue,requestList.request[i]);
      }
      else   /* disk is down : retrieve the parity group to recover */
	retrieveParityGroup(requestList.request[i]);
      break;
    case TEXT:
      if(diskArray.disk[disk].status == UP)
	insertIntoQueue(diskArray.disk[disk].bestEffortQueue,requestList.request[i]);
      else
	retrieveParityGroup(requestList.request[i]);
      break;
    }
  }   
}

void
retrieveParityGroup(Request &request)
{
   int disk,i,ret,startDisk;
   Request req;

             decrementPendingRequestSlot(request); 
   req = request;
   disk = request.disk;
   startDisk = (disk/diskArray.parityGroupSize)*diskArray.parityGroupSize;
   for(i=  startDisk ; i < startDisk + diskArray.parityGroupSize ; i++){
     if(i != disk ){
       req.disk = i;
       switch(request.type){
       case VIDEO:
         if(request.mode==PUSH)
             ret=insertIntoQueue(diskArray.disk[i].periodicQueue, req);
         else
             ret=insertIntoQueue(diskArray.disk[i].aperiodicQueue, req);
         break;
       case TEXT:
         ret=insertIntoQueue(diskArray.disk[i].bestEffortQueue, req);
	 break;
       }
/* requested was inserted : increment pendingSlot */
            if(ret) 
          incrementPendingRequestSlot(request);
	 
     }
   }
}

void
coalesceRequests(RequestList &requestList)
{
  /* coalesce successive requested blocks from a disk as a single block.
      Done mainly to speed up simulations.
      */
      int i,currentReq=requestList.numRequests-2;
      

      while(currentReq >= 0){
	if(requestList.request[currentReq].type == requestList.request[currentReq+1].type &&
	   requestList.request[currentReq].disk == requestList.request[currentReq+1].disk &&
	   requestList.request[ currentReq].blockNo+1 == requestList.request[ currentReq+1].blockNo){
	   requestList.request[currentReq].blockSize += requestList.request[currentReq+1].blockSize;
	   for(i=currentReq+1;i<requestList.numRequests-1;i++)
	     requestList.request[i] = requestList.request[i+1];
	   requestList.numRequests--;
	}
	currentReq--;
      }
}
	  
	  
