/* disk.h -- 
 *  Copyright (c) Distributed Multimedia Computing Laboratory,
 *                The University of Texas at Austin. 1996
 *                All Rights Reserved..
 * Author          : Prashant J. Shenoy
 * Created On      : Tue Apr  2 21:06:55 1996
 * Last Modified By: Prashant J. Shenoy
 * Last Modified On: Tue Jun 24 17:47:50 1997
 * Update Count    : 92
 * Status          : Unknown, Use with caution!
 */

#ifndef _DISK_H
#define _DISK_H

#include "defns.h"

enum raid_level {RAID0=0, RAID1, RAID2, RAID3, RAID4, RAID5};
typedef enum raid_level RaidLevel;

enum file_type {VIDEO,TEXT,AUDIO};
typedef enum file_type FileType;

enum disk_status { UP=0, DOWN };
typedef enum disk_status DiskStatus;

enum schedule_policy {LOOK=0, CLOOK, FCFS,PLOOK,PLOOKEDF};
typedef enum schedule_policy SchedulePolicy;

enum placement_policy{ROUND_ROBIN=0,RANDOM};
typedef enum placement_policy PlacementPolicy;

enum retrieval_mode {PUSH=0,PULL};
typedef enum retrieval_mode RetrievalMode;


struct disk_param{

int bytesSector;
int sectorsTrack;
int tracksCyl;
int numCyl;
double minSeek;
double avgSeek;
double maxSeek;
double rotLatency;
double maxXfrRate;
double A,B,C;
long diskSize;
};

typedef struct disk_param DiskParam;

struct request{
  int client;
  int blockNo;
  int disk;
  int blockSize;
  int flags;
  float arrivalTime;
  float deadline;
  int pendingRequestIndex;
  FileType type;
  RetrievalMode mode;
  double serviceEstimate;
};
typedef struct request Request;

struct request_list{
  Request request[MAX_REQUESTS];
  int numRequests;
};
typedef struct request_list RequestList;

struct request_queue{
  Request queue[MAX_REQUESTS];
  int head;
  int tail;
  FACILITY lock; /* do we need a lock */
  EVENT *notEmpty; /* malloc for sched queue, point to array for others */
  EVENT *notFull;
};
typedef struct request_queue RequestQueue;

struct schedule_stats{
  double idle,latency;
  int data;
  int periodicData,aperiodicData,bestEffortData;
  int scheduler,disk;
};
typedef struct schedule_stats ScheduleStats;

struct disk{
 DiskStatus status;
 RequestQueue periodicQueue;
 RequestQueue aperiodicQueue;
 RequestQueue bestEffortQueue;
 RequestQueue scheduledQueue;
  ScheduleStats scheduleStats; 
  EVENT notEmpty[3];
  EVENT notFull[3];
  EVENT scheduledQNotEmpty;
  EVENT scheduledQNotFull;
  EVENT workConserve;
  int newRoundFlag;
};
typedef struct disk Disk;

struct used_slices{
double periodic,aperiodic,bestEffort;
};
typedef struct used_slices UsedSlices;

struct estimates{
double newReq,nextReq,old;
};
typedef struct estimates Estimates;

struct schedule_param{
  int round;
  double periodicSlice;
  double aperiodicSlice;
  double bestEffortSlice;
  int usePartitioning;
  int partitionTime;
  SchedulePolicy schedulePolicy; /* policy for top level scheduler */
  SchedulePolicy diskPolicy;   /* policy for scheduling the disk */
};
typedef struct schedule_param ScheduleParam;
 
enum block_type{ PARITY, DATA};
typedef enum block_type BlockType;

struct parity_group_table{
  BlockType type;
  int dbt;
};
typedef struct parity_group_table ParityGroupTable;

typedef int DataBlockTable;

enum block_placement{FIXED, VARIABLE};
typedef enum block_placement BlockPlacement;

struct disk_array{
int numDisks;
ScheduleParam scheduleParam;
PlacementPolicy placementPolicy;
Disk *disk;
RaidLevel RAIDLevel;
BlockPlacement blockPlacement;
int blockSize; /* base block size */
int parityGroupSize;
int parityBlockSize;
int stripeDegree;
DiskParam diskParam;
ParityGroupTable **pgt;
DataBlockTable **dbt;
};

typedef struct disk_array DiskArray;


struct disk_location{
 int block;
 int cylinder;
 int sector;
 int track;
 int direction;
};
typedef  struct disk_location DiskLocation;


struct stats{
  double busyTime;
  int video;
  int text;
};
typedef struct stats Stats;


typedef int PendingRequests;

struct statistics{
  Stats *disk;
  PendingRequests **pendingRequests;
  HIST  maxBusyTime, minBusyTime, sumBusyTime;
  HIST maxBlocks, minBlocks, sumBlocks;
  HIST maxVideoBlocks, minVideoBlocks, sumVideoBlocks;
  HIST maxTextBlocks, minTextBlocks, sumTextBlocks;
  HIST textResponseTime,videoResponseTime;
  HIST textQueuingDelay,textServiceTime;
};

typedef struct statistics Statistics;

struct arguments{

char *ifileName;
char *ofileName;
FILE *ifp;
FILE *ofp;

/*  array parameters */
int numDisks;
enum raid_level RAIDLevel;
SchedulePolicy schedulePolicy;
SchedulePolicy diskPolicy;
PlacementPolicy placementPolicy;
int fixedBlockPlacement;
int allowFragmentation;
int blockSize;
int CMblockSize;
int parityGroupSize;
int parityBlockSize;
int stripeDegree;
int serverPush;

/* disk characteristics */
DiskParam diskParam;
  /* disk scheduling characteristics */
int usePartitioning;
int  partitionTime;
double periodicSlice;
double aperiodicSlice;
double bestEffortSlice;
int round;

/* client parameters */
int dataRate;
char traceFile[MAX_STR];
int emptySystem;
double interArrival;
int numCMClients;
int numTextClients;
int numClients;
int requestSize; /* for text */
int pullRequestSize; /* for pull */
int holdTime;


/* file parameters */
int numCMFiles;
int numTextFiles;
int numFiles;
int textFileSize;

/* simulation parameters */
int simulationLength;
int failureRound;
int statsStartTime;
int seed;

  /* output charactristics */
  int verboseOutput;
};


/* ________ define some default disk parameters _______ */
/* ________ Seagate Elite 3 ______________ */
 
const int SE3_bytesSector =  512; 
const int SE3_sectorsTrack=   99; 
const int SE3_tracksCyl   =   21;
const int SE3_numCyl      = 2627;
const int SE3_diskCap     = (2667 * 1024);  /* 2667 MB */
 
const float SE3_minSeek    =  1.7;   /* Single Cylinder Seek time */
const float SE3_avgSeek    = 11.0;   /* All are in ms */
const float SE3_maxSeek    = 22.5;
const float SE3_rotLatency = 11.1;
const float SE3_maxXfrRate =  4.6;   /* MB / sec */

/* _________ globals __________ */
extern DiskArray diskArray;
extern Statistics stats;
extern char err_str[MAX_STR];
#endif


























