/* layout.c --  layout media streams on the array
 *  Copyright (c) Distributed Multimedia Computing Laboratory,
 *                The University of Texas at Austin. 1996
 *                All Rights Reserved..
 * Author          : Prashant J. Shenoy
 * Created On      : Wed Apr  3 21:44:43 1996
 * Last Modified By: Prashant J. Shenoy
 * Last Modified On: Mon Jul  1 20:02:25 1996
 * Update Count    : 93
 * Status          : Unknown, Use with caution!
 */

#include "defns.h"
#include "disk.h"
#include "client.h"
#include "proto.h"

FrameTableElem *frameTable;
FileTableElem *fileTable;
int fileTableSize, frameTableSize;

void
diskLayout(struct arguments &args, DiskArray &diskArray)
{
  initFrameTable(args);
  initFileTable(args);
  LayoutFiles(args,diskArray);
}

void
readTraceFile(char *filename, FrameTableElem * frameTable, int &frameTableSize)
{
  FILE *fp;
  int i;
  
  fp = fopen(filename, "r");
  if (null_ptr(fp))
    error("Unable to open frame sizes file", E_IO);

  for(i=0;i< MAX_FRAMES; i++){
    if(fscanf(fp,"%d",&frameTable[i]) == EOF)
      break;
    // convert bits to bytes
    frameTable[i] /= 8;
  }
  frameTableSize = i ;
}

void
initFrameTable(struct arguments &args)
{
  frameTable = (FrameTableElem *)calloc(MAX_FRAMES,sizeof(FrameTableElem));
  if(null_ptr(frameTable))
    error("Malloc failed in initFrameTable",E_MEM);
  frameTableSize=0;
  readTraceFile(args.traceFile,frameTable,frameTableSize);
  
}

void
initFileTable(struct arguments & args)
{
  int i;
  fileTableSize = args.numFiles;
  fileTable = (FileTableElem *)malloc( sizeof(FileTableElem) * fileTableSize );
  if ( null_ptr(fileTable) ) 
    error("Malloc failed in function initFileTable", E_MEM);
  
  for(i=0; i < args.numCMFiles; i++){
    fileTable[i].type = VIDEO;
    fileTable[i].startfp = random_int(0,frameTableSize-1); /* rand # between 0 and frameTableSize-1*/
    if(args.fixedBlockPlacement)
      fileTable[i].blockSize = args.CMblockSize;
  }
  for(i=args.numCMFiles; i < args.numFiles; i++){
    fileTable[i].type = TEXT;
    fileTable[i].blockSize = args.blockSize;
  }
    
}


void
LayoutFiles(struct arguments &args, DiskArray &diskArray)
{
  Layout(args,diskArray);
  if(diskArray.blockPlacement==VARIABLE && args.allowFragmentation)
    checkLastBlock(args,diskArray); /* make sure last block in each file has appropriate num frames*/
}


void
initLayoutVars(int **currentfp, int **currentbp, int **currentFrameResidue,
	       int **fileList, int **currentDisk, int **firstFrame, int **currentdb, int **permute)
{
  int i;

  *currentfp = (int *) malloc(sizeof(int) * fileTableSize); /* current frame pointer */
  *currentbp = (int *) malloc(sizeof(int) * fileTableSize); /* current Block Pointer */
  *currentFrameResidue = (int *) malloc(sizeof(int) * fileTableSize); /* residue of last frame of prev block in current block */
  *fileList = (int *) malloc(sizeof(int) * fileTableSize);
  *firstFrame = (int *) malloc(sizeof(int) * fileTableSize);
  *currentDisk = (int *) malloc(sizeof(int) * fileTableSize);
  *currentdb = (int *) malloc(sizeof(int)*diskArray.numDisks);
  *permute    = (int *) malloc(sizeof(int)*diskArray.parityGroupSize);

  if(null_ptr(*currentfp) || null_ptr(*currentbp) || null_ptr(*currentFrameResidue) ||
     null_ptr(*fileList)  || null_ptr(*currentDisk) || null_ptr(*currentdb) ||
     null_ptr(*firstFrame) || null_ptr(*permute) )
    error("Malloc failed in initLayoutVars\n",E_MEM);

  for(i=0;i<diskArray.numDisks;i++)
    (*currentdb)[i]=0; /* disk block */

  for(i=0;i<diskArray.parityGroupSize;i++)
    (*permute)[i]=i;  /* identity permutation */

  for(i=0;i<fileTableSize;i++){
    (*currentfp)[i]=0;  /* frame ptr */
    (*currentbp)[i]=0;  /* block ptr */
    (*currentFrameResidue)[i]=0;
    (*currentDisk)[i] = 0;
    (*firstFrame)[i] = fileTable[i].startfp;
    (*fileList)[i] = i; 
  }
}


void
Layout(struct arguments &args, DiskArray &diskArray)
{
  int *currentfp,*currentbp,*currentdb,*currentFrameResidue, *fileList;
  int *currentDisk, *firstFrame, *permute, framesPerRound, fileListSize;
  int i,j,k,m,diskSpaceAvailable,file,numBlocks,bytes, tmpfp,frames;

  initLayoutVars(&currentfp,&currentbp,&currentFrameResidue,&fileList,&currentDisk,
		 &firstFrame, &currentdb, &permute);

  fileListSize = fileTableSize;
  diskSpaceAvailable = 1;

  /* allocate entire parity groups at a time */
  while(diskSpaceAvailable && fileListSize > 0 ){
      file = random_int(0,fileListSize-1); /* pick a file */
      file = fileList[file];
      i = currentDisk[file];
      currentDisk[file] = (currentDisk[file] + diskArray.parityGroupSize) % 
	                   diskArray.numDisks;

      if(diskArray.placementPolicy == RANDOM)
	generatePermutation(permute,diskArray.parityGroupSize);

      for(m=0; m<diskArray.parityGroupSize; m++){
	j = permute[m]; 
	switch(fileTable[file].type){
	case TEXT:
	  numBlocks = fileTable[file].blockSize / diskArray.blockSize;
	  k=0;
	  /* may want to change this : alloc only (G-1) blks not G */
	  while(k<numBlocks){
	    while(isParity(i+j,currentdb[i+j]))
	      currentdb[i+j]++; /* skip parity blocks */
	    if(currentbp[file] * diskArray.blockSize < args.textFileSize && currentbp[file]<MAX_BLOCKS){
	      fileTable[file].inode[ currentbp[file] ].disk = i+j;
	      fileTable[file].inode[ currentbp[file] ].block = currentdb[i+j];
	      fileTable[file].inode[ currentbp[file] ].size = diskArray.blockSize;
#ifdef LAYOUT_DEBUG
	      printf(" file = %3d [text] disk = %3d, block = %6d \n",file,i+j,currentdb[i+j]);
#endif 
	      currentdb[i+j]++;
	      currentbp[file]++;
	      k++;
	    }
	    else{
	      deleteFile(file,fileList,fileListSize);
	      fileTable[file].numBlocks = currentbp[file];
	      break;
	    }
	  }
	 
	  break;
	case VIDEO:
	  if(diskArray.blockPlacement == FIXED)
	    numBlocks = fileTable[file].blockSize / diskArray.blockSize;
	  else{ /* VARIABLE */
	    bytes = currentFrameResidue[file]; /* portion of current frame in this blk*/
	    k=0;
	    framesPerRound = args.dataRate * args.round / 1000;
	    tmpfp = currentfp[file];
	    while(k < framesPerRound  && tmpfp < frameTableSize){
	      bytes += frameTable[(tmpfp + fileTable[file].startfp)%frameTableSize ];
	      tmpfp++;
	      k++;
	    }
	    numBlocks = bytes / (diskArray.blockSize*1024);
	    if(bytes%(diskArray.blockSize*1024)) numBlocks++;
	  }
	  k=frames=0;
	  while(k<numBlocks){
	    while(isParity(i+j,currentdb[i+j]))
	      currentdb[i+j]++;

	    bytes = currentFrameResidue[file]; /* portion of current frame in this blk*/
	    while(bytes < diskArray.blockSize*1024 && 
		  currentfp[file] < frameTableSize){
	      if(args.fixedBlockPlacement || !args.allowFragmentation ||
		 (args.allowFragmentation && frames < framesPerRound)){
		bytes += frameTable[ (currentfp[file] + fileTable[file].startfp) %
				    frameTableSize ];
		currentfp[file]++;
		 frames++;
	      }
	      else 
	       break;
	    }
	    
	    if(currentfp[file] < frameTableSize && currentbp[file] < MAX_BLOCKS){
	      fileTable[file].inode[ currentbp[file] ].disk = i+j;
	      fileTable[file].inode[ currentbp[file] ].block = currentdb[i+j];
	      fileTable[file].inode[ currentbp[file] ].size = diskArray.blockSize;
	      fileTable[file].inode[ currentbp[file] ].residue=
                          currentFrameResidue[file] < diskArray.blockSize*1024 ? currentFrameResidue[file] : -1;
	      fileTable[file].inode[currentbp[file]].firstFrame =  firstFrame[file];
	      fileTable[file].inode[ currentbp[file] ].lastFrame = 
		(currentfp[file] + fileTable[file].startfp - 1 )%frameTableSize;

	      if((fileTable[file].inode[currentbp[file]].lastFrame+1)%frameTableSize == fileTable[file].inode[currentbp[file]].firstFrame )
		fileTable[file].inode[currentbp[file]].firstFrame = (frameTableSize+fileTable[file].inode[currentbp[file]].firstFrame-1)%frameTableSize;	

#ifdef LAYOUT_DEBUG
	      printf(" file = %3d [ cm ] disk = %3d, block = %6d, frames = [%d,%d]\n",file,i+j,currentdb[i+j],fileTable[file].inode[currentbp[file]].firstFrame,fileTable[file].inode[ currentbp[file] ].lastFrame);
#endif
	      currentFrameResidue[file] = bytes >= (diskArray.blockSize*1024) ? 
                      		bytes - (diskArray.blockSize*1024) : -1;
	      if(currentFrameResidue[file] < diskArray.blockSize*1024 )
		firstFrame[file] = (currentfp[file] + fileTable[file].startfp)%frameTableSize;
	      else /* current frame spans entire block and overflows into next blk */
		firstFrame[file] = (frameTableSize+currentfp[file] + fileTable[file].startfp-1)%frameTableSize;
	      

	      currentdb[i+j]++;
	      currentbp[file]++;
	      k++;
	      
	    }
	    else{
	      deleteFile(file,fileList,fileListSize);
	      break;
	    }
	  }
	  break;
	}
	if(currentdb[i+j]* diskArray.blockSize > diskArray.diskParam.diskSize)
	  diskSpaceAvailable=0;
      }
    }
  for(k=0;k<fileTableSize;k++)
    fileTable[k].numBlocks = currentbp[k];  
}




void
deleteFile(int file, int *fileList, int &fileListSize)
{
  int j,i=0;
  while(fileList[i] != file && i < fileListSize)
    i++;
  for(j=i;j<fileListSize;j++)
    fileList[j] = fileList[j+1];
  if(i < fileListSize )
    fileListSize--;
}

void
generatePermutation(int *array, int size)
{
  int i,j,tmp;
  for(i=0;i<size;i++)
    array[i]=i;

  for(i=0;i<size;i++){
    j = random_int(0,size-1);
    /* swap array[i] and array[j] */
    tmp = array[i];
    array[i] = array[j];
    array[j] = tmp;
  }
}
    

/* make sure that the last variable-size block in each file has
 * (roundDuration*dataRate) number of frames; else when a client
 * wraps around the file table, more than 1 blk per round will be accessed
 */
void
checkLastBlock(struct arguments &args,  DiskArray &diskArray)
{
  int i,frames,block,disk,first,last;
  for(i=0;i<fileTableSize;i++){
    if(fileTable[i].type == VIDEO){
      frames=0;
      block=fileTable[i].numBlocks-1;
      disk = fileTable[i].inode[block].disk;
      while(disk == fileTable[i].inode[block].disk){
	first = fileTable[i].inode[block].firstFrame;
	last  = fileTable[i].inode[block].lastFrame;
	if(last>=first)
	  frames += (last-first);
	else /* wrap around in frameTable */
	  frames += (frameTableSize - (first - last));
	if(fileTable[i].inode[block].residue > 0)
	  frames++; 
	if(fileTable[i].inode[(block+1)%fileTable[i].numBlocks].residue == 0)
	  frames++;
	block = (fileTable[i].numBlocks + block - 1 ) % fileTable[i].numBlocks;
      }
      if(frames < args.dataRate * args.round / 1000) 
	fileTable[i].numBlocks = block+1; /* last block does not contain enough frames : delete it */
    }
  }
}
